use std::io::Read;
use std::sync::Arc;

use arrow::datatypes::Schema;
use arrow::record_batch::RecordBatch;
use serde_json::value::Value;

use crate::error::ColumnQError;
use crate::table::{TableLoadOption, TableSchema, TableSource};

fn json_value_from_reader<R: Read>(r: R) -> Result<Value, ColumnQError> {
    serde_json::from_reader(r).map_err(ColumnQError::json_parse)
}

fn json_partition_to_vec(
    json_partition: &Value,
    pointer: Option<&str>,
) -> Result<Vec<Value>, ColumnQError> {
    let mut value_ref = json_partition;

    if let Some(p) = pointer {
        match value_ref.pointer(p) {
            Some(v) => value_ref = v,
            None => {
                return Err(ColumnQError::LoadJson(format!(
                    "Invalid json pointer: {}",
                    p
                )))
            }
        }
    }

    match value_ref.as_array() {
        Some(arr) => Ok(arr.to_vec()),
        None => Err(ColumnQError::LoadJson(format!(
            "{} is not an array",
            pointer.unwrap_or("JSON data")
        ))),
    }
}

fn json_vec_to_partition(
    json_rows: Vec<Value>,
    provided_schema: &Option<TableSchema>,
    batch_size: usize,
    array_encoded: bool,
) -> Result<(arrow::datatypes::Schema, Vec<RecordBatch>), ColumnQError> {
    // load schema
    let schema = match provided_schema {
        Some(s) => s.into(),
        None => arrow::json::reader::infer_json_schema_from_iterator(
            json_rows.iter().map(|v| Ok(v.clone())),
        )
        .map_err(|e| {
            ColumnQError::LoadJson(format!("Failed to infer schema from JSON data: {}", e))
        })?,
    };

    // decode to arrow record batch
    let decoder = arrow::json::reader::Decoder::new(Arc::new(schema.clone()), batch_size, None);
    let batch = {
        // enclose values_iter in its own scope so it won't brrow schema_ref til end of this
        // function
        let mut values_iter: Box<dyn Iterator<Item = arrow::error::Result<Value>>>;
        values_iter = if array_encoded {
            // convert row array to object based on schema
            // TODO: support array_encoded read in upstream arrow json reader instead
            Box::new(json_rows.into_iter().map(|json_row| {
                let mut m = serde_json::map::Map::new();
                schema.fields().iter().enumerate().try_for_each(|(i, f)| {
                    match json_row.get(i) {
                        Some(x) => {
                            m.insert(f.name().to_string(), x.clone());
                            Ok(())
                        }
                        None => Err(arrow::error::ArrowError::JsonError(format!(
                            "arry encoded JSON row missing column {:?} : {:?}",
                            i, json_row
                        ))),
                    }
                })?;
                Ok(Value::Object(m))
            }))
        } else {
            // no need to convert row since each row is already an object
            Box::new(json_rows.into_iter().map(Ok))
        };

        // decode whole array into single record batch
        decoder
            .next_batch(&mut values_iter)
            .map_err(|e| {
                ColumnQError::LoadJson(format!("Failed decode JSON into Arrow record batch: {}", e))
            })?
            .ok_or_else(|| {
                ColumnQError::LoadJson("JSON data results in empty arrow record batch".to_string())
            })?
    };

    Ok((schema, vec![batch]))
}

pub async fn to_mem_table(
    t: &TableSource,
) -> Result<datafusion::datasource::MemTable, ColumnQError> {
    // TODO: make batch size configurable
    let batch_size = 1024;
    let array_encoded = match &t.option {
        Some(TableLoadOption::json { array_encoded, .. }) => array_encoded.unwrap_or(false),
        _ => false,
    };

    if array_encoded && t.schema.is_none() {
        return Err(ColumnQError::LoadJson(
            "Array encoded option requires manually specified schema".to_string(),
        ));
    }

    let pointer = match &t.option {
        Some(TableLoadOption::json { pointer, .. }) => pointer.to_owned(),
        _ => None,
    };

    let mut merged_schema: Option<Schema> = None;
    let json_partitions: Vec<Value> = partitions_from_table_source!(t, json_value_from_reader)?;

    let partitions = json_partitions
        .iter()
        .map(|json_partition| {
            let json_rows = json_partition_to_vec(json_partition, pointer.as_deref())?;
            if json_rows.is_empty() {
                match &pointer {
                    Some(p) => {
                        return Err(ColumnQError::LoadJson(format!(
                            "{} points to an emtpy array",
                            p
                        )));
                    }
                    None => {
                        return Err(ColumnQError::LoadJson(
                            "JSON data is an emtpy array".to_string(),
                        ));
                    }
                }
            }

            let (batch_schema, partition) =
                json_vec_to_partition(json_rows, &t.schema, batch_size, array_encoded)?;

            merged_schema = Some(match &merged_schema {
                Some(s) if s != &batch_schema => Schema::try_merge(vec![s.clone(), batch_schema])?,
                _ => batch_schema,
            });

            Ok(partition)
        })
        .collect::<Result<Vec<Vec<RecordBatch>>, ColumnQError>>()?;

    Ok(datafusion::datasource::MemTable::try_new(
        Arc::new(
            merged_schema
                .ok_or_else(|| ColumnQError::LoadJson("failed to load schema".to_string()))?,
        ),
        partitions,
    )?)
}

#[cfg(test)]
mod tests {
    use super::*;

    use datafusion::datasource::TableProvider;

    use crate::test_util::*;

    #[tokio::test]
    async fn nested_struct_and_lists() -> Result<(), ColumnQError> {
        let t = to_mem_table(&TableSource::new(
            "spacex_launches".to_string(),
            test_data_path("spacex_launches.json"),
        ))
        .await?;

        let schema = t.schema();
        let fields = schema.fields();

        let mut obj_keys = fields.iter().map(|f| f.name()).collect::<Vec<_>>();
        obj_keys.sort();
        let mut expected_obj_keys = vec![
            "fairings",
            "links",
            "static_fire_date_utc",
            "static_fire_date_unix",
            "tbd",
            "net",
            "window",
            "rocket",
            "success",
            "details",
            "crew",
            "ships",
            "capsules",
            "payloads",
            "launchpad",
            "auto_update",
            "failures",
            "flight_number",
            "name",
            "date_unix",
            "date_utc",
            "date_local",
            "date_precision",
            "upcoming",
            "cores",
            "id",
            "launch_library_id",
        ];
        expected_obj_keys.sort();

        assert_eq!(obj_keys, expected_obj_keys);

        Ok(())
    }
}
